# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '4.9.0'
git_version = '927717d1de5a41075133fb61fe82491572a55c27'
full_version = '4.9.0'
is_release = True

if not is_release:
    version = full_version